#!/usr/bin/perl
# author: Fuyuncat
# Site:   WWW.HelloDBA.COM
# Script: http://www.Hellodba.com/Download/OracleNewSort.zip
# Description: Demostrate the new sorting algorithm in Oracle 10g
#

#use strict;
use Data::Dumper;

# data represent in byte, means the range is 0~255
my @data = ( [50, 5], [30, 7], [33, 15], [12, 28], [30, 53], [1, 321], [123, 32], [32, 1], [212, 3], [1, 32]); 

# get minimum number
sub min
{
	if ($_[0] < $_[1]) {
		return $_[0];
	}
	else {
		return $_[1];
	}
}

# sample the QS parition, and return pivot
sub sample
{
	my $arr = shift;
	#return @$arr[rand(@$arr)]; 
	return @$arr[(@$arr)/2]; 
}

# compare item with pivot, ignoring common prefix, 
# diff_at is set to the index of the byte at which the keys of item
# and pivot differ. 
sub compare
{
	my $it = shift;
	my $pv = shift;
	my $cp = shift;
	my $diff_at = shift;

	my $res = 0;
	my $i=$cp;

	# compare item and pivot
	for ($i=$cp; $i<@$it && $i<@$pv; $i++) {
		if (@$it[$i]<@$pv[$i]) {
			$res = -1;
			last;
		}elsif (@$it[$i]>@$pv[$i]) {
			$res = 1;
			last;
		}
	}

	if ($res == 0) {
		$res = @$it - @$pv;
	}

	# set diff_at
	$$diff_at = $i;

	return $res;
}

# link sorted array to result
sub linkResult
{
	my $result = shift;
	my $arr = shift;

	if (defined(@$arr)) {
		push @$result, @$arr;
	}
}

# CPS-QS step for AQS
#
sub AQSdoQS {
	my $items = shift;
	my $common_prefix = shift;

	# partitions for items <, >, ==  to pivot
	my @lt = ();
    my @gt = ();
    my @eq = ();

    my @result = ();

	my $lt_common_prefix = 99999999999999999; # common prefix for items < pivot
	my $gt_common_prefix = 99999999999999999; # common prefix for items > pivot
	my $diff_at, my $compare_res;

	# choose pivot from sample
	my $pivot = sample($items);

	# partitioning data
	foreach my $item(@$items)
	{
		# compare data 
		$compare_res = compare($item, $pivot, $common_prefix, \$diff_at);
		if ($compare_res < 0)	# item < pivot
		{
			push @lt, $item;
			$lt_common_prefix = min($lt_common_prefix, $diff_at);
		}
		elsif ($compare_res == 0) {# item == pivot
			push @eq, $item;
		}
		else	# item > pivot
		{
			push @gt, $item;
			$gt_common_prefix = min($gt_common_prefix, $diff_at);
		}
	}

	# call radix sort 
	if (@lt)
	{
		AQSdoRS(\@lt, $lt_common_prefix);
		linkResult(\@result, \@lt);
	}

	# concatenate equals
	if (@eq) {
	   linkResult(\@result, \@eq);
	}

	# call radix sort 
	if (@gt)
	{
		AQSdoRS(\@gt, $gt_common_prefix);
		linkResult(\@result, \@gt);
	}
	@$items = @result;
}

# radix sort step for AQS
#
sub AQSdoRS
{
	my $items = shift;
	my $common_prefix = shift;

	my @more_buckets = ();
    my @done_buckets = ();
	my @result = (); 

	foreach my $item(@$items)
	{
		if (@$item == $common_prefix)
		{
			push @result, $item;  # key ends before common_prefix
		}
		elsif (@$item == ($common_prefix + 1))
		{ 
			# add item to bucket based on byte at item[common_prefix ]
			push @{$done_buckets[@$item[$common_prefix]]}, $item;
		}
		else
		{ 
			# add item to bucket based on byte at item[common_prefix ]
			push @{$more_buckets[@$item[$common_prefix]]}, $item;
		}
	}
	for (my $i = 0; $i<=255; $i++)
	{
		if (@done_buckets[$i]) {
			linkResult(\@result, @done_buckets[$i]);
		}
		if (@more_buckets[$i]) {
			AQSdoQS(@more_buckets[$i], $ommon_prefix+1);
			linkResult(\@result, @more_buckets[$i]);
		}
	}
	@$items = @result;
}

# print array
sub print_r
{
	my $arr = shift;

	print "[";
	foreach my $item (@$arr) {
		if (scalar(@$item)>1) {
			print_r($item);
		}else {
			print ("\t$item");
		}
	}
	print "]\n";
}

print "unsorted data: \n";
#print Dumper @data;
print_r(\@data);
AQSdoQS(\@data, 0);
print "sorted data: \n";
#print Dumper @data;
print_r(\@data);
